// Copyright 2021 PingCAP, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// See the License for the specific language governing permissions and
// limitations under the License.

package scheduler

import (
	"testing"

	"github.com/stretchr/testify/require"
	"github.com/tikv/migration/cdc/cdc/model"
	"github.com/tikv/migration/cdc/cdc/scheduler/util"
)

// Asserts that BaseSchedulerDispatcher implements InfoProvider interface.
var _ InfoProvider = (*BaseScheduleDispatcher)(nil)

func injectSchedulerStateForInfoProviderTest(dispatcher *BaseScheduleDispatcher) {
	dispatcher.captureStatus = map[model.CaptureID]*captureStatus{
		"capture-1": {
			SyncStatus:   captureSyncFinished,
			CheckpointTs: 1400,
			ResolvedTs:   1500,
		},
		"capture-2": {
			SyncStatus:   captureSyncFinished,
			CheckpointTs: 1300,
			ResolvedTs:   1600,
		},
	}
	dispatcher.keyspans.AddKeySpanRecord(&util.KeySpanRecord{
		KeySpanID: 1,
		CaptureID: "capture-1",
		Status:    util.RunningKeySpan,
	})
	dispatcher.keyspans.AddKeySpanRecord(&util.KeySpanRecord{
		KeySpanID: 2,
		CaptureID: "capture-2",
		Status:    util.RunningKeySpan,
	})
	dispatcher.keyspans.AddKeySpanRecord(&util.KeySpanRecord{
		KeySpanID: 3,
		CaptureID: "capture-1",
		Status:    util.RunningKeySpan,
	})
	dispatcher.keyspans.AddKeySpanRecord(&util.KeySpanRecord{
		KeySpanID: 4,
		CaptureID: "capture-2",
		Status:    util.AddingKeySpan,
	})
	dispatcher.keyspans.AddKeySpanRecord(&util.KeySpanRecord{
		KeySpanID: 5,
		CaptureID: "capture-1",
		Status:    util.RemovingKeySpan,
	})
}

func TestInfoProviderTaskStatus(t *testing.T) {
	dispatcher := NewBaseScheduleDispatcher("cf-1", nil, 1300)
	injectSchedulerStateForInfoProviderTest(dispatcher)

	taskStatus, err := dispatcher.GetTaskStatuses()
	require.NoError(t, err)
	require.Equal(t, map[model.CaptureID]*model.TaskStatus{
		"capture-1": {
			KeySpans: map[model.KeySpanID]*model.KeySpanReplicaInfo{
				1: {},
				3: {},
				5: {},
			},
			Operation: map[model.KeySpanID]*model.KeySpanOperation{
				5: {
					Delete: true,
					Status: model.OperDispatched,
				},
			},
		},
		"capture-2": {
			KeySpans: map[model.KeySpanID]*model.KeySpanReplicaInfo{
				2: {},
				4: {},
			},
			Operation: map[model.KeySpanID]*model.KeySpanOperation{
				4: {
					Delete: false,
					Status: model.OperDispatched,
				},
			},
		},
	}, taskStatus)
}

func TestInfoProviderTaskPosition(t *testing.T) {
	dispatcher := NewBaseScheduleDispatcher("cf-1", nil, 1300)
	injectSchedulerStateForInfoProviderTest(dispatcher)

	taskPosition, err := dispatcher.GetTaskPositions()
	require.NoError(t, err)
	require.Equal(t, map[model.CaptureID]*model.TaskPosition{
		"capture-1": {
			CheckPointTs: 1400,
			ResolvedTs:   1500,
		},
		"capture-2": {
			CheckPointTs: 1300,
			ResolvedTs:   1600,
		},
	}, taskPosition)
}
